import json
import os
import sys
import io
from collections import defaultdict
import numpy as np
from scipy import spatial
from scipy.stats import qmc
import math 
from itertools import product
import parmed
from openmm import Vec3, unit, XmlSerializer, System, CustomNonbondedForce, NonbondedForce, OpenMMException
import openmm.app as app
import openmm.unit as openmmunit
import rdkit
from rdkit import Chem
from rdkit.Chem import AllChem
from openff.toolkit import Molecule, Topology
from openmmforcefields.generators import EspalomaTemplateGenerator, GAFFTemplateGenerator, SMIRNOFFTemplateGenerator
from openmmforcefields.generators.template_generators import SmallMoleculeTemplateGenerator
from cosolvkit.utils import fix_pdb, MutuallyExclusiveParametersError


proteinResidues = ['ALA', 'ASN', 'CYS', 'GLU', 'HIS', 'LEU', 'MET', 'PRO', 'THR', 'TYR', 'ARG', 'ASP', 'GLN', 'GLY', 'ILE', 'LYS', 'PHE', 'SER', 'TRP', 'VAL']
rnaResidues = ['A', 'G', 'C', 'U', 'I']
dnaResidues = ['DA', 'DG', 'DC', 'DT', 'DI']

class CosolventMolecule(object):
    def __init__(self, name: str, smiles: str=None, mol_filename: str=None, resname: str=None, copies: int=None, concentration: float=None):
        """Creates a Cosolvent object.

        :param name: name of the molecule
        :type name: str
        :param smiles: SMILES string of the molecule in the chose protonation state, defaults to None
        :type smiles: str, optional
        :param mol_filename: MOL/SDF filename of the molecule, defaults to None
        :type mol_filename: str, optional
        :param resname: 3-letters residue name of the molecule. If None, the first 3 uppercase letters of the name will be used, defaults to None
        :type resname: str, optional
        :param copies: number of copies of cosolvent the user wants to place, defaults to None
        :type copies: int, optional
        :param concentration: if the number of copies is unknown the user can specify the concentration in Molar units, defaults to None
        :type concentration: float, optional
        """
        if resname is None:
            self.resname = name[:3].upper()
        else:
            self.resname = resname

        # assert not self.resname in AMBER_SUPPORTED_RESNAMES, f"Error: the residue name {self.resname} is already defined in AMBER."
        assert (smiles is not None or mol_filename is not None), "A smiles or a mol/sdf filename has to be defined."
        if mol_filename is not None:
            assert os.path.exists(mol_filename), f"The mol/sdf file {mol_filename} does not exist."

        self.mol = None
        self.atom_names = None
        self.name = name
        self.charge = None
        self.positions = None
        self.mol_filename = None
        self.pdb_conect = None
        self.smiles = smiles
        self.copies = copies
        self.concentration = concentration

        # We use mol file as input for antechamber instead of mol2
        # because of multiple issues I had in the past with that format...
        if mol_filename is None:
            # RDKit parameters for the ETKDGv3 conformational search
            params = AllChem.ETKDGv3()
            params.enforceChirality = True
            params.useSmallRingTorsions = True
            params.useMacrocycleTorsions = True
            params.forceTransAmides = True
            params.useExpTorsionAnglePrefs = True
            params.useBasicKnowledge = True
            # Not a mistake, it's 2 not 3
            # https://www.rdkit.org/docs/RDKit_Book.html#parameters-controlling-conformer-generation
            params.ETVersion = 2

            # Setup rdkit molecule
            mol = Chem.MolFromSmiles(smiles)
            mol.SetProp("_Name", name)
            mol = Chem.AddHs(mol)
            AllChem.EmbedMolecule(mol, params=params)
            AllChem.MMFFOptimizeMolecule(mol)

            mol_filename = '%s.mol' % self.name
            with Chem.SDWriter(mol_filename) as w:
                w.write(mol)
        else:
            mol = Chem.MolFromMolFile(mol_filename, removeHs=False)
            mol.SetProp("_Name", name)

        self.mol = mol
        self.charge = Chem.rdmolops.GetFormalCharge(mol)
        self.atom_names = self._generate_atom_names_from_mol(mol)
        # We make sure the cosolvent is centered, especially if it was provided by the user
        # And honestly, never trust the user!
        positions = mol.GetConformer().GetPositions()
        positions = positions - np.mean(positions, axis=0)
        self.positions = positions
        self.mol_filename = mol_filename
        self.pdb_conect = self._get_pdb_conect(mol)
        return
    
    
    def _generate_atom_names_from_mol(self, rdkit_mol: rdkit.Chem.rdchem.Mol) -> list:
        """Generates atom names from an RDKit molecule.

        :param rdkit_mol: RDKit molecule
        :type rdkit_mol: rdkit.Chem.rdchem.Mol
        :return: list of atom names.
        :rtype: list
        """
        atom_names = []
        counter = {}

        for atom in rdkit_mol.GetAtoms():
            atom_name = atom.GetSymbol()

            if atom_name in counter:
                counter[atom_name] += 1
            else:
                counter[atom_name] = 1

            atom_name += str(counter[atom_name])
            atom_names.append(atom_name)

        return atom_names


    def _get_pdb_conect(self, rdkit_mol: rdkit.Chem.rdchem.Mol) -> list:
        """Generates bonds definition (unused)

        :param rdkit_mol: RDKit molecule
        :type rdkit_mol: rdkit.Chem.rdchem.Mol
        :return: list of bonds
        :rtype: list
        """
        conect = []

        pdb_string = Chem.MolToPDBBlock(rdkit_mol)
        pdb_lines = pdb_string.split('\n')

        for i, line in enumerate(pdb_lines):
            if 'CONECT' in line:
                conect.append(np.array([int(n) for n in line.split()[1:]]))

        return conect

class CosolventSystem(object):
    def __init__(self, 
                 cosolvents: dict,
                 forcefields: dict,
                 simulation_format: str, 
                 modeller: app.Modeller,  
                 padding: openmmunit.Quantity = 12*openmmunit.angstrom, 
                 radius: openmmunit.Quantity = None):
        """Create cosolvent system.

        :param cosolvents: dictionary of cosolvent molecules
        :type cosolvents: dict
        :param forcefields: dictionary of forcefields to use
        :type forcefields: dict
        :param simulation_format: MD format that want to be used for the simulation. Supported formats: Amber, Gromacs, CHARMM, openMM 
        :type simulation_format: str
        :param modeller: openmm modeller created from topology and positions.
        :type modeller: openmm.app.Modeller
        :param padding: specifies the padding used to create the simulation box, defaults to 12*openmmunit.angstrom
        :type padding: openmm.unit.Quantity, optional
        :param radius: Specifies the radius to create the box without receptor, defaults to None
        :type radius: openmm.unit.Quantity, optional
        """
        # Private
        self._available_formats = ["AMBER", "GROMACS", "CHARMM", "OPENMM"]
        self._cosolvent_positions = defaultdict(list)
        self._box = None
        self._periodic_box_vectors = None
        self._box_volume = None
        self._padding = padding

        # Public
        self.protein_radius = 3.5 * openmmunit.angstrom
        self.cosolvents_radius = 2.5*openmmunit.angstrom
        self.modeller = None
        self.system = None
        self.modeller = None
        self.cosolvents = dict()
        self.radius = radius
        
        assert (simulation_format.upper() in self._available_formats), f"Error! The simulation format supplied is not supported! Available simulation engines:\n\t{self._available_formats}"

        # Creating cosolvent molecules
        for c in cosolvents:
            cosolvent = CosolventMolecule(**c)
            cosolvent_xyz = cosolvent.positions*openmmunit.angstrom
            cosolvent_xyz = cosolvent_xyz.value_in_unit(openmmunit.nanometer)
            self.cosolvents[cosolvent] = cosolvent_xyz

        self.modeller = modeller
        
        if self.radius is not None:
            assert (isinstance(radius, openmmunit.Quantity)) and (radius.unit == openmmunit.angstrom), \
                "Error! If no receptor is passed, the radius parameter has to be set and it needs to be in angstrom openmm.unit"
            self.vectors, self.box, self.lowerBound, self.upperBound = self._build_box(None, padding, radius=radius)
            self.receptor = False
        else:
            self.receptor = True
            self.vectors, self.box, self.lowerBound, self.upperBound = self._build_box(self.modeller.positions, padding, radius=None)
        
        # Setting up the box - This has to be done before building the system with
        # the cosolvent molecules.
        self.modeller.topology.setPeriodicBoxVectors(self.vectors)
        self._periodic_box_vectors = self.modeller.topology.getPeriodicBoxVectors().value_in_unit(openmmunit.nanometer)
        vX, vY, vZ = self.modeller.topology.getUnitCellDimensions().value_in_unit(openmmunit.nanometer)
        self.box_volume = vX * vY * vZ
        print("Parametrizing system with forcefields")
        self.forcefield = self._parametrize_system(forcefields, simulation_format, self.cosolvents)
        print(f"Box Volume: {self.box_volume} nm**3")
        print(f"\t{self.box_volume*1000} A^3")
        return
    
#region Public
    def build(self,
              solvent_smiles: str="H2O", 
              n_solvent_molecules: int=None,
              neutralize: bool=True):
        """This function adds thd cosolvents specified in the CosolvSystem
        and solvates with the desired solvent. If n_solvent_molecules is not passed
        the function will try to fill the box with the desired solvent to a certain extent.
        Please note that the solvation with solvents different from water may highly impact
        the execution time.

        :param solvent_smiles: smiles string defining the desired solvent to use, defaults to "H2O"
        :type solvent_smiles: str, optional
        :param n_solvent_molecules: number of molecules of solvent to add, defaults to None
        :type n_solvent_molecules: int, optional
        :param neutralize: if True, the system charge will be neutralized by OpenMM, defaults to True
        :type neutralize: bool, optional
        """
        volume_not_occupied_by_cosolvent = self.fitting_checks()
        assert volume_not_occupied_by_cosolvent is not None, "The requested volume for the cosolvents exceeds the available volume! Please try increasing the box padding or radius."
        receptor_positions = self.modeller.positions.value_in_unit(openmmunit.nanometer)
        cosolv_xyzs = self.add_cosolvents(self.cosolvents, self.vectors, self.lowerBound, self.upperBound, receptor_positions)
        self.modeller = self._setup_new_topology(cosolv_xyzs, self.modeller.topology, self.modeller.positions)
        if solvent_smiles == "H2O":
            if n_solvent_molecules is None: self.modeller.addSolvent(self.forcefield, neutralize=neutralize)
            else: self.modeller.addSolvent(self.forcefield, numAdded=n_solvent_molecules, neutralize=neutralize)
            print(f"Waters added: {self._get_n_waters()}")
        elif solvent_smiles is not None:
            c = {"name": "solvent",
                 "smiles": solvent_smiles}
            solvent_mol = CosolventMolecule(**c)
            cosolv_xyz = solvent_mol.positions*openmmunit.angstrom
            if n_solvent_molecules is not None:
                solvent_mol.copies = n_solvent_molecules
            else:
                one_mol_vol = self.calculate_mol_volume(cosolv_xyz)
                solvent_mol.copies = int(math.floor((volume_not_occupied_by_cosolvent/one_mol_vol)+0.5)*.6)
            d_mol = {solvent_mol: cosolv_xyz.value_in_unit(openmmunit.nanometer)}
            # need to register the custom solvent if not present already
            self.forcefield.registerTemplateGenerator(self._parametrize_cosolvents(d_mol).generator)
            print(f"Placing {solvent_mol.copies}")
            solv_xyz = self.add_cosolvents(d_mol, self.vectors, self.lowerBound, self.upperBound, self.modeller.positions)
            solv_xyz = self.add_cosolvents(d_mol, self.vectors, self.lowerBound, self.upperBound, self.modeller.positions)
            self.modeller = self._setup_new_topology(solv_xyz, self.modeller.topology, self.modeller.positions)
            
        self.system = self._create_system(self.forcefield, self.modeller.topology)
        return
    
    def add_repulsive_forces(self, residues_names: list, epsilon: float=0.01, sigma: float=10.0):
        """This function adds a LJ repulsive potential between the specified molecules.

        :param residues_names: list of residue names
        :type residues_names: list
        :param epsilon: depth of the potential well in kcal/mol, defaults to 0.01
        :type epsilon: float, optional
        :param sigma: inter-particle distance in Angstrom, defaults to 10.0
        :type sigma: float, optional
        """
        epsilon = np.sqrt(epsilon * epsilon) * openmmunit.kilocalories_per_mole
        sigma = sigma * openmmunit.angstrom

        forces = { force.__class__.__name__ : force for force in self.system.getForces()}
        nb_force = forces['NonbondedForce']
        cutoff_distance = nb_force.getCutoffDistance()
        energy_expression = "4*epsilon * (sigma / r)^12;" #Only the repulsive term of the LJ potential
        energy_expression += f"epsilon = {epsilon.value_in_unit_system(openmmunit.md_unit_system)};"
        energy_expression += f"sigma = {sigma.value_in_unit_system(openmmunit.md_unit_system)};"
        repulsive_force = CustomNonbondedForce(energy_expression)
        repulsive_force.addPerParticleParameter("sigma")
        repulsive_force.addPerParticleParameter("epsilon")
        repulsive_force.setNonbondedMethod(NonbondedForce.CutoffPeriodic)
        repulsive_force.setCutoffDistance(cutoff_distance)
        repulsive_force.setUseLongRangeCorrection(False)
        repulsive_force.setUseSwitchingFunction(True)
        repulsive_force.setSwitchingDistance(cutoff_distance - 0.1 * openmmunit.nanometer)

        target_indices = defaultdict(list)
        for i, atom in enumerate(self.modeller.getTopology().atoms()):
            if not atom.residue.name in residues_names:
                charge, sigma, epsilon = nb_force.getParticleParameters(i)
            else:
                target_indices[atom.residue.id].append(i) 
            repulsive_force.addParticle([sigma, epsilon])
        
        for index in range(nb_force.getNumExceptions()):
            idx, jdx, c, s, eps = nb_force.getExceptionParameters(index)
            repulsive_force.addExclusion(idx, jdx)
        
        for res in target_indices.keys():
            indexes = set()
            for x in target_indices.keys():
                if x!= res:
                    for y in target_indices[x]:
                        indexes.add(y)
            repulsive_force.addInteractionGroup(set(target_indices[res]), indexes)
        self.system.addForce(repulsive_force)

        return

    def save_pdb(self, topology: app.Topology, positions: list, out_path: str):
        """Saves the specified topology and position to the out_path file.

        :param topology: topology used
        :type topology: openmm.app.Topology
        :param positions: list of 3D coords
        :type positions: list
        :param out_path: path to where to save the dile
        :type out_path: str
        """
        app.PDBFile.writeFile(
            topology,
            positions,
            open(out_path, "w"),
            keepIds=True
        )
        return

    def save_system(self, out_path: str, system: System):
        """Saves the openmm system to the desired out path.

        :param out_path: path where to save the System
        :type out_path: str
        :param system: system to save
        :type system: openmm.System
        """
        with open(f"{out_path}/system.xml", "w") as fo:
            fo.write(XmlSerializer.serialize(system))
        return
    
    def load_system(self, system_path: str) -> System:
        """Loads the specified openmm system.

        :param system_path: path to the system file.
        :type system_path: str
        :return: a system instance.
        :rtype: openmm.System
        """
        with open(system_path) as fi:
            system = XmlSerializer.deserialize(fi.read())
        return system

    def save_topology(self, topology: app.Topology, positions: list, system: System, simulation_format: str, forcefield: app.ForceField, out_path: str):
        """Save the topology files necessary for MD simulations according to the simulation engine specified.

        :param topology: openmm topology
        :type topology: openmm.app.Topology
        :param positions: list of 3D coordinates
        :type positions: list
        :param system: openmm system
        :type system: openmm.System
        :param simulation_format: name of the simulation engine
        :type simulation_format: str
        :param forcefield: openmm forcefield
        :type forcefield: openmm.app.ForceField
        :param out_path: output path to where to save the topology files
        :type out_path: str
        """
        new_system = forcefield.createSystem(topology,
                                             nonbondedMethod=app.PME,
                                             nonbondedCutoff=10*openmmunit.angstrom,
                                             removeCMMotion=False,
                                             rigidWater=False,
                                             hydrogenMass=3.0*openmmunit.amu)
        
        parmed_structure = parmed.openmm.topsystem.load_topology(topology, new_system, positions)   
        
        simulation_format = simulation_format.upper()
        if simulation_format == "AMBER":
            parmed_structure.save(f'{out_path}/system.prmtop', overwrite=True, format="amber")
            parmed_structure.save(f'{out_path}/system.rst7', overwrite=True, format="rst7")

        elif simulation_format == "GROMACS":
            parmed_structure.save(f'{out_path}/system.top', overwrite=True)
            parmed_structure.save(f'{out_path}/system.gro', overwrite=True)

        elif simulation_format == "CHARMM":
            parmed_structure.save(f'{out_path}/system.psf', overwrite=True)
            parmed_structure.save(f'{out_path}/system.crd', overwrite=True)
            
        elif simulation_format == "OPENMM":
            self.save_system(out_path, system)
            self.save_pdb(topology, positions, f"{out_path}/system.pdb")
            parmed_structure.save(f'{out_path}/system.prmtop', overwrite=True)
        else:
            print("The specified simulation engine is not supported!")
            print(f"Available simulation engines:\n\t{self._available_formats}")
        return
#endregion
    
#region Private
#region Misc
    def _copies_from_concentration(self, water_volume: float):
        """Computes the number of copies of cosolvent necessary to reach the desired concentration

        :param water_volume: volume available to be filled with cosolvents
        :type water_volume: float
        """
        for cosolvent in self.cosolvents:
            if cosolvent.concentration is not None:
                cosolvent.copies = int(math.floor((((cosolvent.concentration*openmmunit.molar)*(water_volume*openmmunit.liters))*openmmunit.AVOGADRO_CONSTANT_NA) + 0.5))
        return

    
    def _get_n_waters(self) -> int:
        """Returns the number of waters in the system.

        :return: number of waters in the system
        :rtype: int
        """
        res = [r.name for r in self.modeller.topology.residues()]
        return res.count('HOH')
    
    def _setup_new_topology(self, cosolvents_positions: dict, receptor_topology: app.Topology = None, receptor_positions:list = None) -> app.Modeller:
        """Returns a new modeller with the topolgy with the new molecules specified

        :param cosolvents_positions: keys are cosolvent molecules and values are lists of position of the new molecules to add
        :type cosolvents_positions: dict
        :param receptor_topology: old topology to which add the new molecules, defaults to None
        :type receptor_topology: openmm.app.Topology, optional
        :param receptor_positions: old positions to which add the new molecules, defaults to None
        :type receptor_positions: list, optional
        :return: new modeller containing combined topology and positions
        :rtype: openmm.app.Modeller
        """
        new_mod = None
        last_res_id = 0
        
        # Add the receptor first if present
        if receptor_topology is not None and receptor_positions is not None and len(receptor_positions) > 0: 
            new_mod = app.Modeller(receptor_topology, receptor_positions)
            last_res_id = int(list(receptor_topology.residues())[-1].index)+1

        # Add cosolvent molecules
        molecules = []
        molecules_positions = []
        cosolvent_names = []
        for cosolvent in cosolvents_positions:
            for i in range(len(cosolvents_positions[cosolvent])):
                cosolvent_names.append(cosolvent.resname)
                mol = Molecule.from_smiles(cosolvent.smiles, name=cosolvent.resname)
                mol.generate_conformers(n_conformers=1)
                molecules.append(mol)
                [molecules_positions.append(x) for x in cosolvents_positions[cosolvent][i]]

        molecules_positions = np.array(molecules_positions)*openmmunit.nanometer
        new_top_openff = Topology.from_molecules(molecules)
        new_top = self._to_openmm_topology(new_top_openff, starting_id=last_res_id)       
        residues = list(new_top.residues())
        for i in range(len(cosolvent_names)):
            residues[i].name = cosolvent_names[i]
          
        if new_mod is None:
            new_mod = app.Modeller(new_top, molecules_positions)
        else:
            new_mod.add(new_top, molecules_positions)
            
        new_mod.topology.setPeriodicBoxVectors(self._periodic_box_vectors)
        return new_mod
    
    def _to_openmm_topology(self, off_topology: Topology, starting_id: int) -> app.Topology:
        """Converts an openff topology to openmm without specifying a different chain for each residue.

        :param off_topology: Openff Topology
        :type off_topology: openff.Topology
        :param starting_id: starting index
        :type starting_id: int
        :raises RuntimeError: if something goes wrong
        :return: openmm topology
        :rtype: openmm.app.Topology
        """
        from openff.toolkit.topology.molecule import Bond

        omm_topology = app.Topology()

        off_topology._ensure_unique_atom_names("residues")

        # Go through atoms in OpenFF to preserve the order.
        omm_atoms = []

        last_chain = None
        cnt = 0
        # For each atom in each molecule, determine which chain/residue it should be a part of
        for molecule in off_topology.molecules:
            # No chain or residue can span more than one OFF molecule, so reset these to None for the first
            # atom in each molecule.
            last_residue = None
            for atom in molecule.atoms:
                # If the residue name is undefined, assume a default of "UNK"
                if "residue_name" in atom.metadata:
                    atom_residue_name = atom.metadata["residue_name"]
                else:
                    atom_residue_name = "UNK"

                # If the residue number is undefined, assume a default of "0"
                if "residue_number" in atom.metadata:
                    atom_residue_number = atom.metadata["residue_number"]
                else:
                    atom_residue_number = str(starting_id+cnt)

                # If the insertion code  is undefined, assume a default of " "
                if "insertion_code" in atom.metadata:
                    atom_insertion_code = atom.metadata["insertion_code"]
                else:
                    atom_insertion_code = " "

                # If the chain ID is undefined, assume a default of "X"
                if "chain_id" in atom.metadata:
                    atom_chain_id = atom.metadata["chain_id"]
                else:
                    atom_chain_id = "X"

                # Determine whether this atom should be part of the last atom's chain, or if it
                # should start a new chain
                if last_chain is None:
                    chain = omm_topology.addChain(atom_chain_id)
                elif last_chain.id == atom_chain_id:
                    chain = last_chain
                else:
                    chain = omm_topology.addChain(atom_chain_id)
                # Determine whether this atom should be a part of the last atom's residue, or if it
                # should start a new residue
                if last_residue is None:
                    residue = omm_topology.addResidue(
                        atom_residue_name,
                        chain,
                        id=atom_residue_number,
                        insertionCode=atom_insertion_code,
                    )
                elif (
                    (last_residue.name == atom_residue_name)
                    and (int(last_residue.id) == int(atom_residue_number))
                    and (last_residue.insertionCode == atom_insertion_code)
                    and (chain.id == last_chain.id)
                ):
                    residue = last_residue
                else:
                    residue = omm_topology.addResidue(
                        atom_residue_name,
                        chain,
                        id=atom_residue_number,
                        insertionCode=atom_insertion_code,
                    )

                # Add atom.
                element = app.Element.getByAtomicNumber(atom.atomic_number)
                omm_atom = omm_topology.addAtom(atom.name, element, residue)

                # Make sure that OpenFF and OpenMM Topology atoms have the same indices.
                assert off_topology.atom_index(atom) == int(omm_atom.id) - 1
                omm_atoms.append(omm_atom)

                last_chain = chain
                last_residue = residue
            
            cnt += 1 
            # Add all bonds.
            bond_types = {1: app.Single, 2: app.Double, 3: app.Triple}
            for bond in molecule.bonds:
                atom1, atom2 = bond.atoms
                atom1_idx, atom2_idx = off_topology.atom_index(
                    atom1
                ), off_topology.atom_index(atom2)
                if isinstance(bond, Bond):
                    if bond.is_aromatic:
                        bond_type = app.Aromatic
                    else:
                        bond_type = bond_types[bond.bond_order]
                    bond_order = bond.bond_order
                elif isinstance(bond, _SimpleBond):
                    bond_type = None
                    bond_order = None
                else:
                    raise RuntimeError(
                        "Unexpected bond type found while iterating over Topology.bonds."
                        f"Found {type(bond)}, allowed are Bond and _SimpleBond."
                    )

                omm_topology.addBond(
                    omm_atoms[atom1_idx],
                    omm_atoms[atom2_idx],
                    type=bond_type,
                    order=bond_order,
                )

        if off_topology.box_vectors is not None:
            from openff.units.openmm import to_openmm

            omm_topology.setPeriodicBoxVectors(to_openmm(off_topology.box_vectors))
        return omm_topology
            
    def _create_system(self, forcefield: app.forcefield, topology: app.Topology) -> System:
        """Returns system created from the Forcefield and the Topology.

        :param forcefield: forcefield(s) used to build the system
        :type forcefield: openmm.app.forcefield
        :param topology: topology used to build the system
        :type topology: openmm.app.Topology
        :return: the new system
        :rtype: openmm.System
        """
        print("Creating system")
        system = forcefield.createSystem(topology,
                                         nonbondedMethod=app.PME,
                                         nonbondedCutoff=10*openmmunit.angstrom,
                                         switchDistance=9*openmmunit.angstrom,
                                         removeCMMotion=True,
                                         constraints=app.HBonds,
                                         hydrogenMass=3.0*openmmunit.amu)
        return system 
#endregion
#region FillTheVoid
    def add_cosolvents(self, 
                       cosolvents: dict, 
                       vectors: tuple[Vec3, Vec3, Vec3], 
                       lowerBound: openmmunit.Quantity | Vec3, 
                       upperBound: openmmunit.Quantity | Vec3,
                       receptor_positions: list) -> dict:
        """This function adds the desired number of cosolvent molecules using the halton sequence
        to generate random uniformly distributed points inside the grid where to place the cosolvent molecules.
        At first, if a receptor/protein is present the halton sequence points that would clash with the protein
        are pruned.

        :param cosolvents: keys are cosolvent molecules and values are 3D coordinates of the molecule
        :type cosolvents: dict
        :param vectors: vectors defining the simulation box
        :type vectors: tuple[openmm.Vec3, openmm.Vec3, openmm.Vec3]
        :param lowerBound: lower bound of the simulation box
        :type lowerBound: openmm.unit.Quantity | Vec3
        :param upperBound: upper bound of the simulation box
        :type upperBound: openmm.unit.Quantity | Vec3
        :param receptor_positions: list of 3D coordinates of the receptor
        :type receptor_positions: list
        :return: keys are cosolvent molecules and values are 3D coordinates of the newly added cosolvent molecules
        :rtype: dict
        """
        edge_cutoff = 2.5*openmmunit.angstrom
        prot_kdtree = None
        # This is used to update the kdtree of the placed cosolvents
        placed_atoms_positions = []
        if receptor_positions is not None and len(receptor_positions) > 0:
            prot_kdtree = spatial.cKDTree(receptor_positions)
        cosolv_xyzs = defaultdict(list)
        
        sampler = qmc.Halton(d=3)
        points = sampler.random(5000000)
        points= qmc.scale(points, [lowerBound[0], 
                                   lowerBound[1], 
                                   lowerBound[2]], 
                                  [upperBound[0], 
                                   upperBound[1], 
                                   upperBound[2]])
        used_halton_ids = list()
        if prot_kdtree is not None:
            banned_ids = prot_kdtree.query_ball_point(points, self.protein_radius.value_in_unit(openmmunit.nanometer))
            used_halton_ids = list(np.unique(np.hstack(banned_ids)).astype(int))
        used_halton_ids = self.delete_edges_points(points, lowerBound, vectors, edge_cutoff.value_in_unit(openmmunit.nanometer), used_halton_ids)
        valid_ids = np.array(range(0, len(points)))
        valid_ids = np.delete(valid_ids, used_halton_ids)
        for cosolvent in cosolvents:
            print(f"Placing {cosolvent.copies} copies of {cosolvent.name}")
            c_xyz = cosolvents[cosolvent]
            while len(cosolv_xyzs[cosolvent]) < cosolvent.copies:
            # for replicate in range(cosolvent.copies):
                # counter = replicate
                if len(placed_atoms_positions) < 1:
                    counter = np.random.choice(len(points))
                    xyz = points[counter]
                    cosolv_xyz = c_xyz + xyz
                    if self.check_coordinates_to_add(cosolv_xyz, None, prot_kdtree):
                        [placed_atoms_positions.append(pos) for pos in cosolv_xyz]
                        cosolv_xyzs[cosolvent].append(cosolv_xyz*openmmunit.nanometer)
                        used_halton_ids.append(counter)
                        kdtree = spatial.cKDTree(placed_atoms_positions)
                else:
                    kdtree = spatial.cKDTree(placed_atoms_positions)
                    cosolv_xyz, valid_ids = self.accept_reject(c_xyz, points, kdtree, valid_ids, lowerBound, vectors, prot_kdtree)

                    if isinstance(cosolv_xyz, int):
                        print("Could not place the cosolvent molecule!")
                    else:
                        cosolv_xyzs[cosolvent].append(cosolv_xyz*openmmunit.nanometer)
                        [placed_atoms_positions.append(pos) for pos in cosolv_xyz]
            print("Done!")
        print("Added cosolvents:")
        for cosolvent in cosolv_xyzs:
            print(f"{cosolvent.name}: {len(cosolv_xyzs[cosolvent])}")
        return cosolv_xyzs

    
    def delete_edges_points(self, xyzs, lowerBound, upperBound, distance, used_halton_ids):
        xyzs = np.atleast_2d(xyzs)
        x, y, z = xyzs[:, 0], xyzs[:, 1], xyzs[:, 2]

        xmin, xmax = lowerBound[0], upperBound[0][0]
        ymin, ymax = lowerBound[1], upperBound[1][1]
        zmin, zmax = lowerBound[2], upperBound[2][2]

        x_close = np.logical_or(np.abs(xmin - x) <= distance, np.abs(xmax - x) <= distance)
        y_close = np.logical_or(np.abs(ymin - y) <= distance, np.abs(ymax - y) <= distance)
        z_close = np.logical_or(np.abs(zmin - z) <= distance, np.abs(zmax - z) <= distance)
        close_to = np.unique(np.argwhere(np.any((x_close, y_close, z_close), axis=0)).flatten())



        return used_halton_ids+list(close_to)

    def check_coordinates_to_add(self, new_coords: np.ndarray, cosolvent_kdtree: spatial.cKDTree, protein_kdtree: spatial.cKDTree) -> bool:
        """Checks that the new coordinates don't clash with the receptor (if present) and/or other cosolvent molecules

        :param new_coords: coordinates of the new molecule of shape (n, 3)
        :type new_coords: np.ndarray
        :param cosolvent_kdtree: tree of the cosolvent molecules present in the box
        :type cosolvent_kdtree: spatial.cKDTree
        :param protein_kdtree: tree of the receptor's coordinates
        :type protein_kdtree: spatial.cKDTree
        :return: True if there are no clashes, False otherwise
        :rtype: bool
        """
        cosolvent_clashes = False
        protein_clashes = False
        check_clashes = cosolvent_kdtree is not None or protein_kdtree is not None
        if cosolvent_kdtree is not None:
            cosolvent_clashes = any(cosolvent_kdtree.query_ball_point(new_coords, self.cosolvents_radius.value_in_unit(openmmunit.nanometer)))
        if protein_kdtree is not None:
            protein_clashes = any(protein_kdtree.query_ball_point(new_coords, self.protein_radius.value_in_unit(openmmunit.nanometer)))
        if check_clashes:
            if not protein_clashes and not cosolvent_clashes:
                return True
            else: return False
        else:
            return self.is_in_box(new_coords, self.lowerBound, self.vectors)

    def accept_reject(self, 
                      xyz: np.ndarray, 
                      halton: list, 
                      kdtree: spatial.cKDTree, 
                      valid_ids: list, 
                      lowerBound: openmmunit.Quantity | Vec3, 
                      upperBound: openmmunit.Quantity | Vec3, 
                      protein_kdtree: spatial.cKDTree) -> tuple[np.ndarray, list]:
        """Accepts or reject the halton move. A random halton point is selected and checked, if accepted
        the cosolvent is placed there, otherwise a local search is performed in the neighbors of the point 
        (1 tile). If the local search produces no clashes the new position is accepted, otherwise a new 
        random halton point is selected and the old one is marked as not good. The algorithm stops
        when a move is accepted or 1000000 of trials are done and no move is accepted.

        :param xyz: 3D coordinates of the cosolvent molecule
        :type xyz: np.ndarray
        :param halton: halton sequence
        :type halton: list
        :param kdtree: tree of the cosolvent molecules positions already placed in the box
        :type kdtree: spatial.cKDTree
        :param valid_ids: valid halton indices
        :type valid_ids: list
        :param lowerBound: lower bound of the box
        :type lowerBound: openmm.unit.Quantity | Vec3
        :param upperBound: upper bound of the box
        :type upperBound: openmm.unit.Quantity | Vec3
        :param protein_kdtree: tree of the protein's positions
        :type protein_kdtree: spatial.cKDTree
        :return: accepted coordinates for the cosolvent and the used halton ids
        :rtype: tuple[np.ndarray, list]
        """        
        trial = 0
        accepted = False
        coords_to_return = 0
        moves = self.local_search()
        while not accepted and trial < 1000000:
            halton_idx = np.random.choice(len(valid_ids))
            rotated_xyz = self.generate_rotation(xyz)
            cosolv_xyz = rotated_xyz + halton[halton_idx]
            valid_ids = np.delete(valid_ids, halton_idx)
            if self.check_coordinates_to_add(cosolv_xyz, kdtree, protein_kdtree):
                accepted = True
                coords_to_return = cosolv_xyz
            else:
                trial += 1
                for move in moves:
                    move = move*openmmunit.angstrom
                    rotated_xyz = self.generate_rotation(xyz)
                    cosolv_xyz = rotated_xyz + halton[halton_idx] + move.value_in_unit(openmmunit.nanometer)
                    if self.is_in_box(cosolv_xyz, lowerBound, upperBound):
                        if self.check_coordinates_to_add(cosolv_xyz, kdtree, protein_kdtree):
                            accepted = True
                            coords_to_return = cosolv_xyz
                            break
                    trial += 1
        return coords_to_return, valid_ids

    def is_in_box(self, 
                  xyzs: np.ndarray, 
                  lowerBound: openmmunit.Quantity | Vec3, 
                  upperBound: openmmunit.Quantity | Vec3) -> bool:
        """Checks if the coordinates are in the box or not

        :param xyzs: coordinates to check
        :type xyzs: np.ndarray
        :param lowerBound: lower bound of the box
        :type lowerBound: openmmunit.Quantity | Vec3
        :param upperBound: upper bound of the box
        :type upperBound: openmmunit.Quantity | Vec3
        :return: True if all the coordinates are in the box, Flase otherwise
        :rtype: bool
        """
        # cutoff = (1.5*openmmunit.angstrom).value_in_unit(openmmunit.nanometer)
        xyzs = np.atleast_2d(xyzs)
        x, y, z = xyzs[:, 0], xyzs[:, 1], xyzs[:, 2]

        xmin, xmax = lowerBound[0], upperBound[0][0]
        ymin, ymax = lowerBound[1], upperBound[1][1]
        zmin, zmax = lowerBound[2], upperBound[2][2]

        x_in = np.logical_and(xmin <= x, x <= xmax)
        y_in = np.logical_and(ymin <= y, y <= ymax)
        z_in = np.logical_and(zmin <= z, z <= zmax)
        all_in = np.all((x_in, y_in, z_in), axis=0)

        return np.all(all_in)

    def local_search(self) -> list:
        """Return all the possible moves in the 1 tile neighbors

        :return: combinations
        :rtype: list
        """
        step = 1
        moves = filter(lambda point: not all(axis ==0 for axis in point), list(product([-step, 0, step], repeat=3)))
        return moves

    def generate_rotation(self, coords: np.ndarray) -> np.ndarray:
        """Rotate a list of 3D [x,y,z] vectors about corresponding random uniformly
            distributed quaternion [w, x, y, z]

        :param coords: list of [x, y, z] cartesian vector coordinates
        :type coords: np.ndarray
        :return: rotated coordinates
        :rtype: np.ndarray
        """
        rand = np.random.rand(3)
        r1 = np.sqrt(1.0 - rand[0])
        r2 = np.sqrt(rand[0])
        pi2 = math.pi * 2.0
        t1 = pi2 * rand[1]
        t2 = pi2 * rand[2]
        qrot = np.array([np.cos(t2) * r2,
                        np.sin(t1) * r1,
                        np.cos(t1) * r1,
                        np.sin(t2) * r2])
        rotation = spatial.transform.Rotation.from_quat(qrot)
        return rotation.apply(coords)
    
#region SizeChecks
    def calculate_mol_volume(self, mol_positions: np.ndarray) -> float:
        """Calculates the volume occupied by the 3D coordinates provided based
        on voxelization.

        :param mol_positions: 3D coordinates of the molecule
        :type mol_positions: np.ndarray
        :return: volume occupied in nm**3
        :rtype: float
        """
        padding = 3.5*openmmunit.angstrom
        offset = 1.5*openmmunit.angstrom
        mesh_step = 0.3*openmmunit.angstrom
        padding = padding.value_in_unit(openmmunit.nanometer)
        offset = offset.value_in_unit(openmmunit.nanometer)
        mesh_step = mesh_step.value_in_unit(openmmunit.nanometer)
        if isinstance(mol_positions, openmmunit.Quantity):
            mol_positions = mol_positions.value_in_unit(openmmunit.nanometer)
        minRange = np.array([min((pos[i] for pos in mol_positions)) for i in range(3)])
        maxRange = np.array([max((pos[i] for pos in mol_positions)) for i in range(3)])
        x = np.arange(minRange[0]-padding, maxRange[0]+padding, mesh_step)
        y = np.arange(minRange[1]-padding, maxRange[1]+padding, mesh_step)
        z = np.arange(minRange[2]-padding, maxRange[2]+padding, mesh_step)
        X, Y, Z = np.meshgrid(x, y, z)
        center_xyzs = np.stack((X.ravel(), Y.ravel(), Z.ravel()), axis=-1)
        kdtree = spatial.cKDTree(center_xyzs)
        query = kdtree.query_ball_point(mol_positions, offset)
        points = np.unique(np.hstack(query)).astype(int)
        return round(len(points)*mesh_step**3, 2)

    def fitting_checks(self) -> float | None:
        """Checks if the required cosolvents can fit in the box and 
        do not exceed the 50% of the available fillable volume 
        (volume not occupied by the receptor, if present).

        :return: available volume if the cosolvents can fit, None otherwise
        :rtype: float | None
        """
        prot_volume = 0
        if self.receptor:
            prot_volume = self.calculate_mol_volume(self.modeller.positions)
            prot_volume = prot_volume
            print(f"Volume protein: {prot_volume*1000} A^3")
        empty_volume = self.cubic_nanometers_to_liters(self.box_volume - prot_volume)
        self._copies_from_concentration(empty_volume)
        cosolvs_volume = defaultdict(float)
        for cosolvent in self.cosolvents:
            cosolvs_volume[cosolvent] = self.calculate_mol_volume(self.cosolvents[cosolvent])*cosolvent.copies
        volume_occupied_by_cosolvent = round(sum(cosolvs_volume.values()), 3)
        empty_available_volume = round(self.liters_to_cubic_nanometers(empty_volume)/2., 3)
        print(f"Volume requested for cosolvents: {volume_occupied_by_cosolvent} nm**3")
        print(f"Volume available for cosolvents: {empty_available_volume} nm**3")
        if volume_occupied_by_cosolvent > empty_available_volume:
            return None
        return empty_available_volume

    def liters_to_cubic_nanometers(self, liters: float | openmmunit.Quantity) -> float:
        """Converts liters in cubic nanometers

        :param liters: volume to convert
        :type liters: float | openmm.unit.Quantity
        :return: converted volume
        :rtype: float
        """
        if isinstance(liters, openmmunit.Quantity):
            liters = liters.value_in_unit(openmmunit.liters)
        value = liters * 1e+24
        return value

    def cubic_nanometers_to_liters(self, vol: float) -> float:
        """Converts cubic nanometers in liters

        :param vol: volume to convert
        :type vol: float
        :return: converted volume
        :rtype: float
        """
        value = vol * 1e-24
        return value
#endregion
#endregion                

#region ForceFieldParametrization
    def _parametrize_system(self, forcefields: dict, engine: str, cosolvents: dict) -> app.ForceField:
        """Parametrize the system with the specified forcefields

        :param forcefields: dictionary of the forcefields to use (from forcefields.json)
        :type forcefields: dict
        :param engine: name of the simulation engine to use
        :type engine: str
        :param cosolvents: cosolvent moleucles (from cosolvents.json)
        :type cosolvents: dict
        :return: forcefield object
        :rtype: openmm.app.ForceField
        """
        engine = engine.upper()
        forcefield = app.ForceField(*forcefields[engine])
        sm_ff = forcefields["small_molecules"][0]
        small_molecule_ff = self._parametrize_cosolvents(cosolvents, small_molecule_ff=sm_ff)
        forcefield.registerTemplateGenerator(small_molecule_ff.generator)
        return forcefield

    def _parametrize_cosolvents(self, cosolvents: dict, small_molecule_ff="espaloma") -> SmallMoleculeTemplateGenerator:
        """Parametrizes cosolvent molecules according to the forcefiled specified.

        :param cosolvents: cosolvents specified
        :type cosolvents: dict
        :param small_molecule_ff: name of the forcefield to use, defaults to "espaloma"
        :type small_molecule_ff: str, optional
        :return: forcefield object for the small molecules
        :rtype: SmallMoleculeTemplateGenerator
        """
        molecules = list()
        for cosolvent in cosolvents:
            try:
                molecules.append(Molecule.from_smiles(cosolvent.smiles, name=cosolvent.name))
            except Exception as e:
                print(e)
                print(cosolvent)
        if small_molecule_ff == "espaloma":
            small_ff = EspalomaTemplateGenerator(molecules=molecules, forcefield='espaloma-0.3.2', template_generator_kwargs={"reference_forcefield": "openff_unconstrained-2.1.0", "charge_method": "nn"})
        elif small_molecule_ff == "gaff":
            small_ff = GAFFTemplateGenerator(molecules=molecules, forcefield='gaff-2.11')
        else:
            small_ff = SMIRNOFFTemplateGenerator(molecules=molecules)
        return small_ff
#endregion
    
#region SimulationBox
    def _build_box(self, 
                   positions: np.ndarray, 
                   padding: openmmunit.Quantity, 
                   radius: openmmunit.Quantity = None) -> tuple[tuple[Vec3, Vec3, Vec3], 
                                                                Vec3, 
                                                                openmmunit.Quantity | Vec3,
                                                                openmmunit.Quantity | Vec3]:
        """Builds the simulation box. If a receptor is passed it is used alongside with the padding
        parameter to build the box automatically, otherwise a radius has to be passed. If no receptor
        the box is centered on the point [0, 0, 0].

        :param positions: coordinates of the receptor if present
        :type positions: np.ndarray
        :param padding: padding to be used
        :type padding: openmm.unit.Quantity
        :param radius: radius specified if no receptor is passed, defaults to None
        :type radius: openmm.unit.Quantity, optional
        :return: The first element returned is a tuple containing the three vectors describing the simulation box.
                The second element is the box itself.
                Third and fourth elements are the lower and upper bound of the simulation box.
        :rtype: tuple[tuple[Vec3, Vec3, Vec3], Vec3, openmmunit.Quantity | Vec3, openmmunit.Quantity | Vec3]
        """
        padding = padding.value_in_unit(openmmunit.nanometer)
        if positions is not None:
            positions = positions.value_in_unit(openmmunit.nanometer)
            minRange = Vec3(*(min((pos[i] for pos in positions)) for i in range(3)))
            maxRange = Vec3(*(max((pos[i] for pos in positions)) for i in range(3)))
            center = 0.5*(minRange+maxRange)
            radius = max(unit.norm(center-pos) for pos in positions)
        else:
            center = Vec3(0, 0, 0)
            radius = radius.value_in_unit(openmmunit.nanometer)
            maxRange = Vec3(radius, radius, radius)
            minRange = Vec3(-radius, -radius, -radius)
        width = max(2*radius+padding, 2*padding)
        vectors = (Vec3(width, 0, 0), Vec3(0, width, 0), Vec3(0, 0, width))
        box = Vec3(vectors[0][0], vectors[1][1], vectors[2][2])
        origin = center - (np.ceil(np.array((width, width, width))))
        self._box_origin = origin
        self._box_size = np.ceil(np.array([maxRange[0]-minRange[0],
                                           maxRange[1]-minRange[1],
                                           maxRange[2]-minRange[2]])).astype(int)
        lowerBound = center-box/2
        upperBound = center+box/2
        return vectors, box, lowerBound, upperBound
#endregion
#endregion
    
class CosolventMembraneSystem(CosolventSystem):  
    def __init__(self, 
                 cosolvents: str,
                 forcefields: str,
                 simulation_format: str, 
                 modeller: app.Modeller,  
                 padding: openmmunit.Quantity = 12*openmmunit.angstrom, 
                 radius: openmmunit.Quantity = None,
                 lipid_type: str=None,
                 lipid_patch_path: str=None):
        """Creates a CosolventMembraneSystem.

        :param cosolvents: path to the cosolvents.json file
        :type cosolvents: str
        :param forcefields: path to the forcefields.json file
        :type forcefields: str
        :param simulation_format: MD format that want to be used for the simulation
        :type simulation_format: str
        :param modeller: Modeller containing topology and positions information
        :type modeller: openmm.app.Modeller
        :param padding: specify the padding to be used to create the simulation box, defaults to 12*openmmunit.angstrom
        :type padding: openmm.unit.Quantity, optional
        :param radius: specifies the radius to create the box without receptor, defaults to None
        :type radius: openmm.unit.Quantity, optional
        :param lipid_type: lipid type to use to build the membrane system, defaults to None. Supported types: ["POPC", "POPE", "DLPC", "DLPE", "DMPC", "DOPC", "DPPC"]. 
                                        Mutually exclusive with <lipid_patch_path>.
        :type lipid_type: str, optional
        :param lipid_patch_path: if lipid type is None the path to a pre-equilibrated patch of custom lipids membrane can be passed, defaults to None. Mutually exclusive with <lipid_type>
        :type lipid_patch_path: str, optional
        :raises MutuallyExclusiveParametersError: custom Exception
        """
        super().__init__(cosolvents=cosolvents,
                         forcefields=forcefields,
                         simulation_format=simulation_format,
                         modeller=modeller,
                         padding=padding,
                         radius=radius)
        
        self.protein_raidus = 1.5 * openmmunit.angstrom
        self.cosolvents_radius = 2.5*openmmunit.angstrom           
        self.lipid_type = lipid_type
        self.lipid_patch = None
        
        self._available_lipids = ["POPC", "POPE", "DLPC", "DLPE", "DMPC", "DOPC", "DPPC"]
        self._cosolvent_placement = None
         
        if self.lipid_type is not None and lipid_patch_path is None:
            assert self.lipid_type in self._available_lipids, print(f"Error! The specified lipid is not supported! Please choose between the following lipid types:\n\t{self._available_lipids}")
        elif lipid_patch_path is not None and self.lipid_type is None:
            self.lipid_patch = app.PDBFile(lipid_patch_path)
        else:
            raise MutuallyExclusiveParametersError("Error! <lipid_type> and <lipid_patch_path> are mutually exclusive parameters. Please pass just one of them.")
    
    def add_membrane(self, cosolvent_placement: int=0, neutralize: bool=True, waters_to_keep: list=None):
        """Create a membrane system

        :param cosolvent_placement: determines on what side of the membrane will the cosolvents be placed, defaults to 0.
                                    * -1: inside the membrane
                                    * +1: outside the membrane
                                    * 0: everywhere
        :type cosolvent_placement: int, optional
        :param neutralize: if neutralize the system when solvating the membrane, defaults to True
        :type neutralize: bool, optional
        :param waters_to_keep: a list of the indices of key waters that should not be deleted, defaults to None
        :type waters_to_keep: list, optional
        :raises SystemError: if OpenMM is not able to relax the system after adding the membrane a SystemError is raised
        """
        waters_residue_names = ["HOH", "WAT"]
        # OpenMM default
        padding = 1 * openmmunit.nanometer
        self._cosolvent_placement = cosolvent_placement
        if self._cosolvent_placement == 0: print("No preference of what side of the membrane to place the cosolvents")
        elif self._cosolvent_placement == 1: print("Placing cosolvent molecules outside of the membrane")
        elif self._cosolvent_placement == -1: print("Placing cosolvent molecules inside the membrane")
        else: 
            print("Error! Available options for <cosolvent_placement> are [0 -> no preference, 1 -> outside, -1 -> inside]")
            raise SystemError
        try:
            if self.lipid_type is not None:
                self.modeller.addMembrane(forcefield=self.forcefield,
                                        lipidType=self.lipid_type,
                                        neutralize=neutralize,
                                        minimumPadding=padding)
            elif self.lipid_patch is not None:
                self.modeller.addMembrane(forcefield=self.forcefield,
                                        lipidType=self.lipid_patch,
                                        neutralize=neutralize,
                                        minimumPadding=padding)
            if waters_to_keep is not None:
                waters_to_delete = [atom for atom in self.modeller.topology.atoms() if atom.residue.index not in waters_to_keep and atom.residue.name in waters_residue_names]
            else:
                waters_to_delete = [atom for atom in self.modeller.topology.atoms() if atom.residue.name in waters_residue_names]
            self.modeller.delete(waters_to_delete)
        except OpenMMException as e:
            print("Something went wrong during the relaxation of the membrane.\nProbably a problem related to particle's coordinates.")
            sys.exit(1)
        print("Membrane system built.")
        positions = self.modeller.positions.value_in_unit(openmmunit.nanometer)
        minRange = Vec3(*(min((pos[i] for pos in positions)) for i in range(3)))
        maxRange = Vec3(*(max((pos[i] for pos in positions)) for i in range(3)))
        center = 0.5*(minRange+maxRange)
        radius = max(unit.norm(center-pos) for pos in positions)
        width = 2*radius
        self.vectors = self.modeller.topology.getPeriodicBoxVectors().value_in_unit(openmmunit.nanometer)
        box = Vec3(self.vectors[0][0], self.vectors[1][1], self.vectors[2][2])
        origin = center - (np.ceil(np.array((width, width, width))))
        self._box_origin = origin
        self._box_size = np.ceil(np.array([maxRange[0]-minRange[0],
                                           maxRange[1]-minRange[1],
                                           maxRange[2]-minRange[2]])).astype(int)
        self.lowerBound = center-box/2
        self.upperBound = center+box/2
        self._periodic_box_vectors = self.modeller.topology.getPeriodicBoxVectors().value_in_unit(openmmunit.nanometer)
        return

    def build(self, neutralize: bool=True):
        """Adds the cosolvent molecules to the system

        :param neutralize: if neutralize the system during solvation, defaults to True
        :type neutralize: bool, optional
        """
        if self._cosolvent_placement != 0:
            lipid_positions = list()
            atoms = list(self.modeller.topology.atoms())
            positions = self.modeller.positions.value_in_unit(openmmunit.nanometer)
            for i in range(len(atoms)):
                if atoms[i].residue.name not in proteinResidues and atoms[i].residue.name not in dnaResidues and atoms[i].residue.name not in rnaResidues:
                    lipid_positions.append(positions[i])
            minRange = min((pos[2] for pos in lipid_positions))
            maxRange = max((pos[2] for pos in lipid_positions))
            if self._cosolvent_placement == -1:
                upperBound = Vec3(self.upperBound[0], self.upperBound[1], minRange)
                lowerBound = self.lowerBound
            else:
                upperBound = self.upperBound
                lowerBound = Vec3(self.lowerBound[0], self.lowerBound[1],maxRange)
        else:
            upperBound = self.upperBound
            lowerBound = self.lowerBound
        print("Checking volumes...")
        volume_not_occupied_by_cosolvent = self.fitting_checks()
        assert volume_not_occupied_by_cosolvent is not None, "The requested volume for the cosolvents exceeds the available volume! Please try increasing the box padding or radius."
        receptor_positions = self.modeller.positions.value_in_unit(openmmunit.nanometer)
        cosolv_xyzs = self.add_cosolvents(self.cosolvents, self.vectors, lowerBound, upperBound, receptor_positions)
        self.modeller = self._setup_new_topology(cosolv_xyzs, self.modeller.topology, self.modeller.positions)
        self.modeller.addSolvent(forcefield=self.forcefield, neutralize=neutralize)
            
        self.system = self._create_system(self.forcefield, self.modeller.topology)
        return
        



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# CoSolvKit
#
# Class to analyze cosolvent MD
#

import os
import sys
import json
import numpy as np
from scipy import spatial
from scipy.ndimage import gaussian_filter
from scipy.signal import correlate
from scipy.interpolate import RegularGridInterpolator
from gridData import Grid
import MDAnalysis as mda
from MDAnalysis import Universe
from MDAnalysis.analysis import rdf
from MDAnalysis.analysis.base import AnalysisBase
import matplotlib.pyplot as plt
import matplotlib.style as style
import pandas as pd
import pymol
from pymol import cmd, stored
from cosolvkit.cosolvent_system import CosolventMolecule


BOLTZMANN_CONSTANT_KB = 0.0019872041


def _normalization(data, a=0, b=0):
    """_summary_

    :param data: list of data points
    :type data: list
    :param a: int a, defaults to 0
    :type a: int, optional
    :param b: int b, defaults to 0
    :type b: int, optional
    :return: normalized data
    :rtype: list
    """
    min_data = np.min(data)
    max_data = np.max(data)
    return a + ((data - min_data) * (b - a)) / (max_data - min_data)


def _smooth_grid_free_energy(gfe, sigma=1):
    """_summary_

    :param gfe: _description_
    :type gfe: _type_
    :param sigma: _description_, defaults to 1
    :type sigma: int, optional
    :return: _description_
    :rtype: _type_
    """
    # We keep only the favorable spots with energy < 0 kcal/mol
    gfe_fav = np.copy(gfe)
    gfe_fav[gfe > 0.] = 0.
    # We smooth the energy a bit
    gfe_fav_smooth = gaussian_filter(gfe_fav, sigma=sigma)
    # Normalize data between the original energy minima value and 0
    gfe_smooth_norm = _normalization(gfe_fav_smooth, np.min(gfe_fav), np.max(gfe_fav))
    # Put the favorable smoothed data in the original grid
    gfe[gfe_smooth_norm < 0] = gfe_smooth_norm[gfe_smooth_norm < 0]

    return gfe


def _grid_free_energy(hist, volume_water, gridsize, n_atoms, n_frames, temperature=300.):
    # Avoid 0 in the histogram for the log function
    hist = hist + 1E-20
    # The volume here is the volume of water and not the entire box
    volume_voxel = gridsize **3
    n_voxel = volume_water / volume_voxel
    # Probability of the solute in the bulk (without protein)
    N_o = n_atoms / n_voxel
    # Probability of the solute (with the protein)
    N = hist / n_frames
    # Atomic grid free energy
    gfe = -(BOLTZMANN_CONSTANT_KB * temperature) * np.log(N / N_o)

    return gfe


def _grid_density(hist):
    return (hist - np.mean(hist)) / np.std(hist)


def _subset_grid(grid, center, box_size, gridsize=0.5):
    # Create grid interpolator
    # Number of midpoints is equal to the number of grid points
    grid_interpn = RegularGridInterpolator(grid.midpoints, grid.grid)

    # Create sub grid coordinates
    # We get first the edges of the grid box, and after the midpoints
    # So this we are sure (I guess) that the sub grid is well centered on center
    # There might be a better way of doing this... Actually I tried, but didn't worked very well.
    x, y, z = center
    sd = box_size / 2.
    hbins = np.round(box_size / gridsize).astype(int)
    edges = (np.linspace(0, box_size[0], num=hbins[0] + 1, endpoint=True) + (x - sd[0]),
             np.linspace(0, box_size[1], num=hbins[1] + 1, endpoint=True) + (y - sd[1]),
             np.linspace(0, box_size[2], num=hbins[2] + 1, endpoint=True) + (z - sd[2]))
    midpoints = (edges[0][:-1] + np.diff(edges[0]) / 2.,
                 edges[1][:-1] + np.diff(edges[1]) / 2.,
                 edges[2][:-1] + np.diff(edges[2]) / 2.)
    X, Y, Z = np.meshgrid(midpoints[0], midpoints[1], midpoints[2])
    xyzs = np.stack((X.ravel(), Y.ravel(), Z.ravel()), axis=-1)
    # Configuration of the sub grid
    origin_subgrid = (midpoints[0][0], midpoints[1][0], midpoints[2][0])
    shape_subgrid = (midpoints[0].shape[0], midpoints[1].shape[0], midpoints[2].shape[0])

    # Do interpolation
    sub_grid_values = grid_interpn(xyzs)
    sub_grid_values = sub_grid_values.reshape(shape_subgrid)
    sub_grid_values = np.swapaxes(sub_grid_values, 0, 1)
    sub_grid = Grid(sub_grid_values, origin=origin_subgrid, delta=gridsize)

    return sub_grid


def _export(fname, grid, gridsize=0.5, center=None, box_size=None):
    assert (center is None and box_size is None) or (center is not None and box_size is not None), \
           "Both center and box size have to be defined, or none of them."

    if center is None and box_size is None:
        grid.export(fname)
    elif center is not None and box_size is not None:
        center = np.array(center)
        box_size = np.array(box_size)

        assert np.ravel(center).size == 3, "Error: center should contain only (x, y, z)."
        assert np.ravel(box_size).size == 3, "Error: grid size should contain only (a, b, c)."
        assert (box_size > 0).all(), "Error: grid size cannot contain negative numbers."

        sub_grid = _subset_grid(grid, center, box_size, gridsize)
        sub_grid.export(fname)


class Analysis(AnalysisBase):
    """Analysis class to generate density grids

    :param AnalysisBase: Base MDAnalysis class
    :type AnalysisBase: AnalysisBase
    """
    def __init__(self, atomgroup, gridsize=0.5, **kwargs):
        super(Analysis, self).__init__(atomgroup.universe.trajectory, **kwargs)

        if atomgroup.n_atoms == 0:
            print("Error: no atoms were selected.")
            sys.exit(1)

        self._u = atomgroup.universe
        self._ag = atomgroup
        self._gridsize = gridsize
        self._nframes = 0
        self._n_atoms = atomgroup.n_atoms
        self._center = None
        self._box_size = None

    def _prepare(self):
        self._positions = []
        self._centers = []
        self._dimensions = []

    def _single_frame(self):
        self._positions.append(self._ag.atoms.positions.astype(float))
        self._dimensions.append(self._u.dimensions[:3])
        self._centers.append(self._u.atoms.center_of_geometry())
        self._nframes += 1

    def _conclude(self):
        self._positions = np.array(self._positions, dtype=float)
        self._box_size = np.mean(self._dimensions, axis=0)
        self._center = np.mean(self._centers, axis=0)

        # Get all the positions
        positions = self._get_positions()

        # Get grid edges and origin
        x, y, z = self._center
        sd = self._box_size / 2.
        hbins = np.round(self._box_size / self._gridsize).astype(int)
        self._edges = (np.linspace(0, self._box_size[0], num=hbins[0] + 1, endpoint=True) + (x - sd[0]),
                       np.linspace(0, self._box_size[1], num=hbins[1] + 1, endpoint=True) + (y - sd[1]),
                       np.linspace(0, self._box_size[2], num=hbins[2] + 1, endpoint=True) + (z - sd[2]))
        origin = (self._edges[0][0], self._edges[1][0], self._edges[2][0])

        hist, edges = np.histogramdd(positions, bins=self._edges)
        self._histogram = Grid(hist, origin=origin, delta=self._gridsize)
        self._density = Grid(_grid_density(hist), origin=origin, delta=self._gridsize)

    def _get_positions(self, start=0, stop=None):
        positions = self._positions[start:stop,:,:]
        new_shape = (positions.shape[0] * positions.shape[1], 3)
        positions = positions.reshape(new_shape)
        return positions

    def atomic_grid_free_energy(self, volume, temperature=300., atom_radius=1.4, smoothing=True):
        """Compute grid free energy.
        """
        agfe = _grid_free_energy(self._histogram.grid, volume, self._gridsize, self._n_atoms, self._nframes, temperature)

        if smoothing:
            # We divide by 3 in order to have radius == 3 sigma
            agfe = _smooth_grid_free_energy(agfe, atom_radius / 3.)

        self._agfe = Grid(agfe, edges=self._histogram.edges)

    def export_histogram(self, fname, gridsize=0.5, center=None, box_size=None):
        """ Export histogram maps
        """
        _export(fname, self._histogram, gridsize, center, box_size)

    def export_density(self, fname, gridsize=0.5, center=None, box_size=None):
        """ Export density maps
        """
        _export(fname, self._density, gridsize, center, box_size)

    def export_atomic_grid_free_energy(self, fname, gridsize=0.5, center=None, box_size=None):
        """ Export atomic grid free energy
        """
        _export(fname, self._agfe, gridsize, center, box_size)

class Report:
    """Report class. This is the main class that takes care of post MD simulation processing and analysis.
    """
    def __init__(self, log_file, traj_file, top_file, cosolvents_path):
        """_summary_

        :param log_file: log file generated by MD Simulation. In CosolvKit this is called statistics.csv
        :type log_file: str
        :param traj_file: Trajectory file generated by MD Simulation.
        :type traj_file: str
        :param top_file: Topology file generated by CosolvKit.
        :type top_file: str
        :param cosolvents_path: path to the json file defining the cosolvents present in the system.
        :type cosolvents_path: str
        """
        self.statistics = log_file
        self.trajectory = traj_file
        self.topology = top_file
        self.universe = Universe(self.topology, self.trajectory)
        self.cosolvents = list()

        with open(cosolvents_path) as fi:
            cosolvents_d = json.load(fi)
        for cosolvent in cosolvents_d:
            self.cosolvents.append(CosolventMolecule(**cosolvent))
        
        self._volume = None
        self._temperature = None
        self._potential_energy = None
        self._potential_energy, self._temperature, self._volume = self._get_temp_vol_pot(self.statistics)
        return
    
    def generate_report(self, out_path):
        """Creates the main plots for RDFs, autocorrelations and equilibration.

        :param out_path: path to where to save the results. 
        :type out_path: str
        """
        print("Generating report...")
        # setup results folders
        report_path = os.path.join(out_path, "report")
        rdf_path = os.path.join(report_path, "rdf")
        os.makedirs(report_path, exist_ok=True)
        os.makedirs(rdf_path, exist_ok=True)

        # Generate equilibration plot
        self._plot_temp_vol_pot(report_path)
        print("Plotting RDFs")
        self._rdf_mda(self.universe, self.cosolvents, rdf_path)
        return
    
    def generate_density_maps(self, out_path, analysis_selection_string=""):
        """Generates the desnity maps for the target cosolvents.

        :param out_path: path to where to save the density files.
        :type out_path: str
        :param analysis_selection_string: MD Analysis selection string if want to generate densities only for specific molecules, defaults to ""
        :type analysis_selection_string: str, optional
        """
        print("Generating density maps...")
        os.makedirs(out_path, exist_ok=True)
        volume = self._volume[-1]
        temperature = self._temperature[-1]
        if analysis_selection_string == "":
            print("No cosolvent specified for the densities analysis. Generating a density map for each cosolvent.")
            for cosolvent in self.cosolvents:
                selection_string = f"resname {cosolvent.resname}"
                self._run_analysis(selection_string=selection_string,
                                   volume=volume,
                                   temperature=temperature,
                                   out_path=out_path,
                                   cosolvent_name=cosolvent.resname)
        else:
            print(f"Generating density maps for the following selection string: {analysis_selection_string}")
            self._run_analysis(selection_string=analysis_selection_string, 
                               volume=volume,
                               temperature=temperature,
                               out_path=out_path,
                               cosolvent_name=None)
        return
    
    def generate_pymol_reports(self, topology, trajectory, density_files, selection_string, out_path):
        """Generate the PyMol reports from the density maps.

        :param topology: Topology generated by CosolvKit.
        :type topology: str
        :param trajectory: Trajectory generated by MD Simulation.
        :type trajectory: str
        :param density_files: list of density files to include in the same PyMol session. Limited to 5.
        :type density_files: list
        :param selection_string: PyMol selection string if willing to specify target residues.
        :type selection_string: str
        :param out_path: path to where to save the PyMol session.
        :type out_path: str
        """
        colors = ['marine', 
                  'orange', 
                  'magenta',
                  'salmon',
                  'purple']
        
        assert len(density_files) < len(colors), "Error! Too many density files, not enough colors available!"
        base_path = os.getcwd()
        topology = os.path.join(base_path, topology)
        trajectory = os.path.join(base_path, trajectory)
        # density_file = os.path.join(base_path, density_file)
        
        cmd_string = ""
        # Load topology and first frame of the trajectory
        cmd.load(topology, "structure")
        cmd.load_traj(trajectory, start=0, stop=1)
        cmd_string += f"cmd.load('{topology}', 'structure')\n"
        cmd_string += f"cmd.load_traj('{trajectory}', start=0, stop=1)\n"

        # Load density
        for idx, density in enumerate(density_files):
            cosolv = density.split('_')[-1]

            cmd.load(density, f"density_map_{cosolv}")
            cmd_string += f"cmd.load('{density}', 'density_map_{cosolv}')\n"

        # Set structure's color
        cmd.color("grey50", "structure and name C*")

        # Remove solvent and organic molecules
        cmd.remove("solvent")
        cmd.remove("org")
        cmd_string += f"cmd.color('grey50', 'structure and name C*')\n"
        cmd_string += f"cmd.remove('solvent')\n"
        cmd_string += f"cmd.remove('org')\n"

        for idx in range(len(density_files)):
            # Create isomesh for hydrogen bond probes
            cmd.isomesh(f"dens_{cosolv}", f"density_map_{cosolv}", 10)

            # Color the hydrgen bond isomesh
            cmd.color(colors[idx], f"dens_{cosolv}")
            cmd_string += f"cmd.isomesh('dens_{cosolv}', 'density_map_{cosolv}', 10)\n"
            cmd_string += f"cmd.color('{colors[idx]}', 'dens_{cosolv}')\n"
            
        # Show sticks for the residues of interest
        if selection_string != '':
            cmd.show("sticks", selection_string)
            cmd_string += f"cmd.show('sticks', '{selection_string}')\n"

        cmd.hide("spheres")
        # Set valence to 0 - no double bonds
        cmd.set("valence", 0)
        # Set cartoon_side_chain_helper to 1 - less messy
        cmd.set("cartoon_side_chain_helper", 1)
        # Set background color
        cmd.bg_color("grey80")
        cmd_string += "cmd.hide('spheres')\n"
        cmd_string += "cmd.set('valence', 0)\n"
        cmd_string += "cmd.set('cartoon_side_chain_helper', 1)\n"
        cmd_string += "cmd.bg_color('grey80')"
        
        with open(os.path.join(out_path, "pymol_session_cmd.pml"), "w") as fo:
            fo.write(cmd_string)
            
        cmd.save(os.path.join(out_path, "pymol_results_session.pse"))
        return
    
    def _run_analysis(self, selection_string, volume, temperature, out_path, cosolvent_name=None):
        """Creates Analysis object and generates densities.

        :param selection_string: MD Analysis selection string.
        :type selection_string: str
        :param volume: volume of the system.
        :type volume: float
        :param temperature: temperature of the system.
        :type temperature: float
        :param out_path: path to where to save the results.
        :type out_path: str
        :param cosolvent_name: name of the cosolvent if not analysing all the cosolvents in the system, defaults to None
        :type cosolvent_name: str, optional
        """
        fig_density_name = os.path.join(out_path, f"map_density.dx")
        fig_energy_name =  os.path.join(out_path, f"map_agfe.dx")
        if cosolvent_name is not None:
            fig_density_name = os.path.join(out_path, f"map_density_{cosolvent_name}.dx")
            fig_energy_name =  os.path.join(out_path, f"map_agfe_{cosolvent_name}.dx")
        analysis = Analysis(self.universe.select_atoms(selection_string), verbose=True)
        analysis.run()
        analysis.atomic_grid_free_energy(volume, temperature)
        analysis.export_density(fig_density_name)
        analysis.export_atomic_grid_free_energy(fig_energy_name)
        self.density_file = fig_density_name
        return

    def _get_temp_vol_pot(self, log_file):
        """Returns temperature, volume and potential energy of the system during the MD simulation.

        :param log_file: log file generated by the MD simulation. In CosolvKit is statistics.csv.
        :type log_file: str
        :return: potential energy, temperature and volume of the system for each frame.
        :rtype: tuple(list, list, list)
        """
        df = pd.read_csv(log_file)
        pot_e = list(df["Potential Energy (kJ/mole)"])
        temp = list(df["Temperature (K)"])
        vol = list(df["Box Volume (nm^3)"])
        return pot_e, temp, vol

    def _plot_temp_vol_pot(self, outpath=None):
        """Plots equilibration data.

        :param outpath: path to where to save the plot, defaults to None
        :type outpath: str, optional
        """
        if outpath is not None:
            fig_name = f"{outpath}/simulation_statistics.png"

        fig, axs = plt.subplots(3, 1, figsize=(12, 6))

        axs[0].plot(self._potential_energy, color='green', linewidth=2)
        axs[0].set_title('Potential Energy',)
        axs[0].set_xlabel('Time (ps)')
        axs[0].set_ylabel('Energy (kJ/mole)')
    
        axs[1].plot(self._volume, color='blue', linewidth=2)
        axs[1].set_title('Volume')
        axs[1].set_xlabel('Time (ps)')
        axs[1].set_ylabel('Volume (nm^3)')

        axs[2].plot(self._temperature, color='red', linewidth=2)
        axs[2].set_title('Temperature')
        axs[2].set_xlabel('Time (ps)')
        axs[2].set_ylabel('Temperature (K)')

        plt.tight_layout()
        plt.savefig(fig_name)
        plt.close()

        return 
    
    def _rdf_mda(self, universe: Universe, cosolvents: list, outpath=None):
        """Generates the plots for RDFs and Autocorrelations.

        :param universe: MD Analysis Universe that is created from the topology and trajectories.
        :type universe: Universe
        :param cosolvents: list of cosolvents in the system
        :type cosolvents: list
        :param outpath: path to where to save the plots, defaults to None
        :type outpath: str, optional
        """
        np.seterr(divide='ignore', invalid='ignore')
        wat_resname = "HOH"
        # if top.endswith("cosolv_system.prmtop"):
        #     wat_resname = "WAT"
        oxygen_atoms = universe.select_atoms(f"resname {wat_resname} and name O")
        sim_frames = len(universe.trajectory)
        step_size = int(sim_frames/250)
        if step_size < 1:
            step_size = 1
        n_bins = 150
        for cosolvent in cosolvents:
            cosolvent_name = cosolvent.resname
            r_max = 15
                
            cosolvent_residues = universe.select_atoms(f'resname {cosolvent_name}')
            atoms_names = cosolvent_residues.residues[0].atoms.names
            for cosolvent_atom in set(atoms_names):
                max_y = 0
                if "H" in cosolvent_atom: continue
                print(f"Analysing {cosolvent_name}-{cosolvent_atom}")
                fig, ax = plt.subplots(2, 2, sharex=False, sharey=False)
                plt.tight_layout(pad=3.0)
                # Here compute RDF between same atoms and different molecules
                atoms = cosolvent_residues.select_atoms(f'name {cosolvent_atom}')
                irdf = rdf.InterRDF(atoms, atoms, nbins=n_bins, range=(0.0, r_max), exclusion_block=(1, 1))
                irdf.run(start=0, step=step_size)
                max_y = max(irdf.results.rdf)
                ax[0][0].plot(irdf.results.bins, irdf.results.rdf, label="RDF")
                ax[0][0].set_xlabel(r'$r$ $\AA$')
                ax[0][0].set_ylabel("$g(r)$")
                ax[0][0].set_title(f"RDF-{cosolvent_name} {cosolvent_atom}")
                # ax[0][0].set_title(f"RDF-{cosolvent_name} {cosolvent_atom} every {step_size} frames")
                leg = ax[0][0].legend(handlelength=0, handletextpad=0, fancybox=True)
                for item in leg.legendHandles:
                    item.set_visible(False)
                
                ax[1][0] = self._plot_autocorrelation(data=irdf.results.rdf,
                                                        ax=ax[1][0], 
                                            cosolvent_name1=cosolvent_name, 
                                            cosolvent_atom1=cosolvent_atom, 
                                            cosolvent_name2=cosolvent_name, 
                                            cosolvent_atom2=cosolvent_atom)
                # Here compute RDF between atom and water's oxygen
                irdf = rdf.InterRDF(atoms, oxygen_atoms, nbins=n_bins, range=(0.0, r_max))
                irdf.run(start=0, step=step_size)
                max_y = max(irdf.results.rdf)
                irdf.run()
                ax[0][1].plot(irdf.results.bins, irdf.results.rdf, label="RDF")
                ax[0][1].set_xlabel(r'$r$ $\AA$')
                ax[0][1].set_ylabel("$g(r)$")
                ax[0][1].set_title(f"RDF {cosolvent_name} {cosolvent_atom}-HOH O")
                # ax[0][1].set_title(f"RDF {cosolvent_name} {cosolvent_atom}-HOH O every {step_size} frames")
                leg = ax[0][1].legend(handlelength=0, handletextpad=0, fancybox=True)
                for item in leg.legendHandles:
                    item.set_visible(False)

                self._plot_autocorrelation(data=irdf.results.rdf, 
                                             ax=ax[1][1], 
                                             cosolvent_name1=cosolvent_name, 
                                             cosolvent_atom1=cosolvent_atom, 
                                             cosolvent_name2="HOH", 
                                             cosolvent_atom2="O")
                if outpath is not None:
                    plt.savefig(f"{outpath}/rdf_{cosolvent_name}_{cosolvent_atom}.png")
                plt.close()
        
        # Finally do waters
        print("Analysing water")
        r_max = 8.5
        fig, ax = plt.subplots()
        plt.setp(ax, xlim=(0, r_max+1))
        irdf = rdf.InterRDF(oxygen_atoms, oxygen_atoms, nbins=n_bins, range=(0.0, r_max), exclusion_block=(1, 1))
        irdf.run(start=0, step=50)
        # irdf.run()
        ax.plot(irdf.results.bins, irdf.results.rdf, label="RDF")
        ax.set_xlabel(r'$r$ $\AA$')
        ax.set_ylabel("$g(r)$")
        ax.set_title(f"RDF-HOH O every 50 frames")
        leg = ax.legend(handlelength=0, handletextpad=0, fancybox=True)
        for item in leg.legendHandles:
            item.set_visible(False)
        if outpath is not None:
            plt.savefig(f"{outpath}/rdf_HOH_O.png")
        plt.close()
        return

    def _autocorrelation(self, data):        
        """Gets the autocorrelation values.

        :param data: list of data for which the autocorrelation has to be computed.
        :type data: list
        :return: list of autocorrelations.
        :rtype: list
        """
        n = len(data)
        mean = np.mean(data)
        autocorr = correlate(data - mean, data - mean, mode='full', method='auto')
        return autocorr[n - 1:]
    
    def _plot_autocorrelation(self, data, ax, cosolvent_name1=None, cosolvent_atom1=None, cosolvent_name2=None, cosolvent_atom2=None):
        """Plots autocorrelations.

        :param data: list of data points for which we want to plot autocorrelations.
        :type data: list
        :param ax: matplotlib axis to add the autocorrelation to the RDF plot.
        :type ax: matplotlib.pyplot.axisß
        :param cosolvent_name1: name of the first cosolvent molecule, defaults to None
        :type cosolvent_name1: str, optional
        :param cosolvent_atom1: name of the first atom, defaults to None
        :type cosolvent_atom1: str, optional
        :param cosolvent_name2: name of the second cosolvent molecule, defaults to None
        :type cosolvent_name2: str, optional
        :param cosolvent_atom2: name of the second atom, defaults to None
        :type cosolvent_atom2: str, optional
        :return: the axis with the autocorrelation plot
        :rtype: matplotlib.pyplot.axis
        """
        title = f"{cosolvent_name1} {cosolvent_atom1}-{cosolvent_name2} {cosolvent_atom2}"
        data = data[0::2]
        autocorr_values = self._autocorrelation(data)
        # Normalize autocorrelation values for better plotting
        normalized_autocorr = autocorr_values / np.max(np.abs(autocorr_values))
        lags = np.arange(0, len(autocorr_values))
        pd.plotting.autocorrelation_plot(pd.Series(normalized_autocorr), ax=ax, label="Autocorrelation")
        ax.grid(False)
        ax.set_xlim([0, len(autocorr_values)])
        ax.set_title(title)
        ax.set_xlabel('Lag')
        ax.set_ylabel('Autocorrelation')
        leg = ax.legend(handlelength=0, handletextpad=0, fancybox=True)
        for item in leg.legendHandles:
            item.set_visible(False)
        return ax


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# CoSolvKit
#

from .cosolvent_system import CosolventSystem, CosolventMolecule
from .analysis import Report

__all__ = ["CosolventSystem", "CosolventMolecule", "Report"]



import os
import io
import json
import time
import argparse
from collections import defaultdict
from cosolvkit.config import Config
from cosolvkit.utils import fix_pdb, add_variants
from cosolvkit.cosolvent_system import CosolventSystem, CosolventMembraneSystem
from cosolvkit.simulation import run_simulation
from openmm.app import *
from openmm import *
import openmm.unit as openmmunit

def cmd_lineparser():
    parser = argparse.ArgumentParser(description="Runs cosolvkit and MD simulation afterwards.",
                                     epilog="""
        REPORTING BUGS
                Please report bugs to:
                AutoDock mailing list   http://autodock.scripps.edu/mailing_list\n

        COPYRIGHT
                Copyright (C) 2023 Forli Lab, Center for Computational Structural Biology,
                             Scripps Research.""")
    
    parser.add_argument('-c', '--config', dest='config', required=True,
                        action='store', help='path to the json config file')
    return parser.parse_args()

def main():
    args = cmd_lineparser()
    config_file = args.config
    config = Config.from_config(config_file)
    # Start setting up the pipeline
    os.makedirs(config.output, exist_ok=True)
    
    if config.run_cosovlent_system:
        if (config.receptor and config.radius is not None) or (not config.receptor and config.radius is None):
            raise SystemExit("Error! If the config file specifies a receptor, the radius should be set to null and vice versa.")
        
        # Check if need to clean the protein and add variants of reisudes
        if config.receptor:
            print("Protein present")
            with open(config.protein_path) as f:
                pdb_string = f.read()
            pdb_string = io.StringIO(pdb_string)
            if config.clean_protein:
                pdbfile = None
                pdbxfile = None
                if config.protein_path.endswith(".pdb"):
                    pdbfile = pdb_string
                else:
                    pdbxfile = pdb_string
                protein_topology, protein_positions = fix_pdb(pdbfile=pdbfile,
                                                            pdbxfile=pdbxfile, 
                                                            keep_heterogens=config.keep_heterogens, )
            else:
                if not config.protein_path.endswith(".pdb"):
                    pdb = PDBxFile(pdb_string)
                else:
                    pdb = PDBFile(pdb_string)
                protein_topology, protein_positions = pdb.topology, pdb.positions
            
            # Call add_variants fucntion
            if len(config.variants_d.keys()) > 0:
                variants = list()
                residues = list(protein_topology.residues())
                mapping = defaultdict(list)
                for r in residues:
                    mapping[r.chain.id].append(int(r.id))
                
                for chain in mapping:
                    for res_number in mapping[chain]:
                        key = f"{chain}:{res_number}"
                        if key in config.variants_d:
                            variants.append(config.variants_d[key])
                        else:
                            variants.append(None)
                protein_topology, protein_positions = add_variants(protein_topology, protein_positions, variants)
                
        else:
            assert config.radius is not None, "radius is None in the config"
            # Create empty modeller since there's nothing in the system yet
            config.radius = config.radius * openmmunit.angstrom
            protein_topology, protein_positions = Topology(), None

        protein_modeller = Modeller(protein_topology, protein_positions)

        # Load cosolvents and forcefields dictionaries
        with open(config.cosolvents) as fi:
            cosolvents = json.load(fi)

        with open(config.forcefields) as fi:
            forcefields = json.load(fi)

        print("Building cosolvent system")
        if config.membrane:
            cosolv_system = CosolventMembraneSystem(cosolvents=cosolvents,
                                                    forcefields=forcefields,
                                                    simulation_format=config.md_format,
                                                    modeller=protein_modeller,
                                                    lipid_type=config.lipid_type,
                                                    lipid_patch_path=config.lipid_patch_path)
            cosolv_system.add_membrane(cosolvent_placement=config.cosolvent_placement,
                                    waters_to_keep=config.waters_to_keep)
            cosolv_system.build()
        else:
            cosolv_system = CosolventSystem(cosolvents=cosolvents,
                                            forcefields=forcefields,
                                            simulation_format=config.md_format,
                                            modeller=protein_modeller,
                                            radius=config.radius)
            cosolv_system.build(solvent_smiles=config.solvent_smiles,
                                n_solvent_molecules=config.solvent_copies)
            
        if config.add_repulsive:
            cosolv_system.add_repulsive_forces(config.repulsive_residues, epsilon=config.epsilon, sigma=config.sigma)

        print("Saving topology file")
        cosolv_system.save_topology(topology=cosolv_system.modeller.topology, 
                                    positions=cosolv_system.modeller.positions,
                                    system=cosolv_system.system,
                                    simulation_format=config.md_format,
                                    forcefield=cosolv_system.forcefield,
                                    out_path=config.output)
    
    if config.run_md:
        print("Running MD simulation")
        start = time.time()
        if config.md_format.upper() != "OPENMM":
            # Change the next two lines depending on the simulation_format you chose
            topo = os.path.join(config.output, "system.prmtop")
            pos = os.path.join(config.output, "system.rst7")
            # This is for openmm
            pdb = None
            system = None
        else:
            topo = None
            pos = None
            # This is for openmm
            pdb = os.path.join(config.output, "system.pdb")
            system = os.path.join(config.output, "system.xml")
        
        if config.md_format.upper() == "OPENMM":
            print(f"Starting MD simulation from the files: {pdb}, {system}")
        else:
            print(f"Starting MD simulation from the files: {topo}, {pos}")
        
        run_simulation(
                        simulation_format = config.md_format,
                        topology = topo,
                        positions = pos,
                        pdb = pdb,
                        system = system,
                        warming_steps = 100000,
                        simulation_steps = 6250000, # 25ns
                        results_path = config.output, # This should be the name of system being simulated
                        seed=None
        )
        print(f"Simulation finished after {(time.time() - start)/60:.2f} min.")
    return


if __name__ == "__main__":
    sys.exit(main())


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# CoSolvKit
#
# Utils functions
#

import contextlib
import os
import tempfile
import shutil
import subprocess
import importlib
from collections import defaultdict

import numpy as np
import mdtraj
from MDAnalysis import Universe
from openmm.app import *
from openmm import *
import pdbfixer

class MutuallyExclusiveParametersError(Exception):
    """A custom exception.

    :param Exception: this is a custom exception for mutually exclusive parameters
    :type Exception: Exception
    """
    pass   


# def add_harmonic_restraints(prmtop, inpcrd, system, atom_selection, k=1.0):
#     """Add harmonic restraints to the system.

#     Args:
#         prmtop (AmberPrmtopFile): Amber topology object
#         inpcrd (AmberInpcrdFile): Amber coordinates object
#         system (System): OpenMM system object created from the Amber topology object
#         atom_selection (str): Atom selection (see MDTraj documentation)
#         k (float): harmonic force restraints in kcal/mol/A**2 (default: 1 kcal/mol/A**2)

#     Returns:
#         (list, int): list of all the atom ids on which am harmonic force is applied, index with the System of the force that was added

#     """
#     mdtop = mdtraj.Topology.from_openmm(prmtop.topology)
#     atom_idxs = mdtop.select(atom_selection)
#     positions = inpcrd.positions

#     # Tranform constant to the right unit
#     k = k * unit.kilocalories_per_mole / unit.angstroms**2
#     k = k.value_in_unit_system(unit.md_unit_system)

#     if atom_idxs.size == 0:
#         print("Warning: no atoms selected using: %s" % atom_selection)
#         return ([], None)

#     # Take into accoun the periodic condition
#     # http://docs.openmm.org/latest/api-python/generated/simtk.openmm.openmm.CustomExternalForce.html
#     force = CustomExternalForce("k * periodicdistance(x, y, z, x0, y0, z0)^2")
    
#     harmonic_force_idx = system.addForce(force)
    
#     force.addGlobalParameter("k", k)
#     print(force.getGlobalParameterDefaultValue(0))
#     force.addPerParticleParameter("x0")
#     force.addPerParticleParameter("y0")
#     force.addPerParticleParameter("z0")
    
#     for atom_idx in atom_idxs:
#         #print(atom_idx, positions[atom_idx].value_in_unit_system(units.md_unit_system))
#         force.addParticle(int(atom_idx), positions[atom_idx].value_in_unit_system(unit.md_unit_system))

#     return atom_idxs


# def update_harmonic_restraints(simulation, k=1.0):
#     """Update harmonic restraints force
    
#     Args:
#         simulation (Simulation): OpenMM simulation object
#         k (float): new harmonic force constraint in kcal/mol/A**2 (default: 1 kcal/mol/A**2)

#     """
#     # Tranform constant to the right unit
#     k = k * unit.kilocalories_per_mole / unit.angstroms**2
#     k = k.value_in_unit_system(unit.md_unit_system)
    
#     simulation.context.setParameter('k', k)

def fix_pdb(pdbfile: str, pdbxfile: str, keep_heterogens: bool=False) -> tuple[Topology, list]:
    """Fixes common problems in PDB such as:
            - missing atoms
            - missing residues
            - missing hydrogens
            - remove nonstandard residues

    :param pdbfile: pdb string old format
    :type pdbfile: str
    :param pdbxfile: pdb string new format
    :type pdbxfile: str
    :param keep_heterogens: if False all heterogen atoms but waters are deleted, defaults to False
    :type keep_heterogens: bool, optional
    :return: new topology and positions
    :rtype: tuple[Topology, list]
    """
    fixer = pdbfixer.PDBFixer(pdbfile=pdbfile, pdbxfile=pdbxfile)
    fixer.findMissingResidues()
    
    chains = list(fixer.topology.chains())
    keys = fixer.missingResidues.keys()
    for key in list(keys):
        chain = chains[key[0]]
        if key[1] == 0 or key[1] == len(list(chain.residues())):
            del fixer.missingResidues[key]

    if not keep_heterogens:
        fixer.removeHeterogens(keepWater=True)

    fixer.findMissingAtoms() 
    fixer.addMissingAtoms()
    fixer.addMissingHydrogens(7)
    return fixer.topology, fixer.positions
    
def add_variants(topology: Topology, positions: list, variants: list=list()) -> tuple[Topology, list]:
    """Adds variants for specific protonation states.

    :param topology: openmm topology
    :type topology: Topology
    :param positions: openmm positions
    :type positions: list
    :param variants: list of variants to apply for the protonation states, defaults to list()
    :type variants: list, optional
    :return: topology and positions with added protonation states
    :rtype: tuple[Topology, list]
    """
    modeller = Modeller(topology, positions)
    added_variants = modeller.addHydrogens(variants=variants)
    return modeller.topology, modeller.positions



import sys
import argparse
from cosolvkit.analysis import Report

def cmd_lineparser():
    parser = argparse.ArgumentParser(description="Runs cosolvkit and MD simulation afterwards.")
    parser.add_argument('-l', '--log_file', dest='log_file', required=True,
                        action='store', help='path to the log file from MD simulation')
    parser.add_argument('-traj', dest='traj_file', required=True,
                        action='store', help='path to the traj <.dcd> file from MD simulation')
    parser.add_argument('-topo', dest='top_file', required=True,
                        action='store', help='path to the topology file from MD simulation')
    parser.add_argument('-densitises', nargs='+', help='path to the density files passed as a list',
                        action='store', dest='densities', required=True)
    parser.add_argument('-o', '--out_path', dest='out_path', required=True,
                        action='store', help='path where to store output plots')
    parser.add_argument('-c', '--cosolvents', dest='cosolvents', required=True,
                        action='store', help='path to the json file containing cosolvents used for the simulation')
    return parser.parse_args()

def main():
    args = cmd_lineparser()
    log_file = args.log_file
    traj_file = args.traj_file
    top_file = args.top_file
    densities = args.densities
    out_path = args.out_path
    cosolvents_path = args.cosolvents

    report = Report(log_file, traj_file, top_file, cosolvents_path)
    report.generate_report(out_path=out_path)
    report.generate_pymol_reports(report.topology, 
                                  report.trajectory, 
                                  density_files=densities, 
                                  selection_string='', 
                                  out_path=out_path)
    return


if __name__ == "__main__":
    sys.exit(main())


import os
import json
from inspect import signature

class Config(object):
    """This class handles the config.json options and the whole Cosolvent setup.

    :param object: inherits from the base class
    :type object: object
    """
    def __init__(self, 
                 cosolvents=None,
                 forcefields=None,
                 md_format=None,
                 receptor=True,
                 protein_path=None,
                 clean_protein=True,
                 keep_heterogens=False,
                 variants_d=dict(),
                 add_repulsive=False,
                 repulsive_residues=list(),
                 epsilon=None,
                 sigma=None,
                 solvent_smiles=None,
                 solvent_copies=None,
                 membrane=False,
                 lipid_type=None,
                 lipid_patch_path=None,
                 cosolvent_placement=0,
                 waters_to_keep=list(),
                 radius=None,
                 output=None,
                 run_cosolvent_system=True,
                 run_md=False):
        
        self.cosolvents = cosolvents
        self.forcefields = forcefields
        self.md_format = md_format
        self.receptor = receptor
        self.protein_path = protein_path
        self.clean_protein = clean_protein
        self.keep_heterogens = keep_heterogens
        self.variants_d = variants_d
        self.add_repulsive = add_repulsive
        self.repulsive_residues = repulsive_residues
        self.epsilon = epsilon
        self.sigma = sigma
        self.solvent_smiles = solvent_smiles
        self.solvent_copies = solvent_copies
        self.membrane = membrane
        self.lipid_type = lipid_type
        self.lipid_patch_path = lipid_patch_path
        self.cosolvent_placement = cosolvent_placement
        self.waters_to_keep = waters_to_keep
        self.radius = radius
        self.output = output
        self.run_cosovlent_system = run_cosolvent_system
        self.run_md = run_md
    
    @classmethod
    def get_defaults_dict(cls):
        """Returns a dictionary of all the class attributes

        :return: dictionary of class attributes
        :rtype: dict
        """
        defaults = {}
        sig = signature(cls)
        for key in sig.parameters:
            defaults[key] = sig.parameters[key].default 
        return defaults

        
    @classmethod
    def from_config(cls, config):
        """Sets up the parameters to run cosolvent from the config.json file supplied.

        :param config: loads the config.json file and populates the class attributes
        :type config: str
        :raises ValueError: raises an error if some attributes are not recognized
        :return: instance of the Config class
        :rtype: Config
        """
        expected_keys = cls.get_defaults_dict().keys()
        with open(config) as f:
            config = json.load(f)
        bad_keys = [k for k in config if k not in expected_keys]
        if len(bad_keys) > 0:
            err_msg = "unexpected keys in Config.from_config():" + os.linesep
            for key in bad_keys:
                err_msg += "  - %s" % key + os.linesep
            raise ValueError(err_msg)
        p = cls(**config)
        return p


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import glob
import fnmatch
from setuptools import setup, find_packages


def find_files(directory):
    matches = []

    for root, dirnames, filenames in os.walk(directory):
        for filename in fnmatch.filter(filenames, '*'):
            matches.append(os.path.join(root, filename))

    return matches


setup(name="cosolvkit",
      version='0.4.4',
      description="CosolvKit",
      author="Niccolo Bruciaferri, Jerome Eberhardt",
      author_email="forli@scripps.edu",
      url="https://github.com/forlilab/cosolvkit",
      packages=find_packages(exclude=['docs']),
      package_data={"cosolvkit" : ["data/*"]},
      data_files=[("", ["README.md", "LICENSE"])],
      include_package_data=True,
      zip_safe=False,
      python_requires=">=3.9",
      license="LGPL-2.1",
      keywords=["molecular modeling", "drug design",
                "cosolvent", "MD simulations"],
      classifiers=["Programming Language :: Python",
                   "Operating System :: Unix",
                   "Operating System :: MacOS",
                   "Topic :: Scientific/Engineering"],
      entry_points={
          'console_scripts': [
              'create_cosolvent_system=scripts.create_cosolvent_system:main',
              'post_simulation_processing=scripts.post_simulation_processing:main'
          ]
      }
      
)



import os
from sys import stdout
import openmm.app as app
import openmm
import openmm.unit as openmmunit
from mdtraj.reporters import NetCDFReporter, DCDReporter

def run_simulation( simulation_format: str = 'OPENMM',
                    results_path: str = "output",
                    topology: str=None, 
                    positions: str=None,
                    pdb: str = 'output/system.pdb',
                    system: str = 'output/system.xml', 
                    warming_steps: int = 100000,
                    simulation_steps: int = 25000000,
                    seed: int = None
                    ):
    """_summary_

    :param simulation_format: determines what MD engine will be used, defaults to 'OPENMM'. Available engines are: [AMBER, GROMACS, CHARMM, OPENMM]
    :type simulation_format: str, optional
    :param results_path: path to where to save the results, defaults to "output"
    :type results_path: str, optional
    :param topology: path to the topology file if using simulation_format different from OPENMM, defaults to None
    :type topology: str, optional
    :param positions: path to the positions file if using simulation_format different from OPENMM, defaults to None
    :type positions: str, optional
    :param pdb: path to the pdb file if using simulation_format OPENMM, defaults to 'output/system.pdb'
    :type pdb: str, optional
    :param system: path to the system.xml file if using simulation_format OPENMM, defaults to 'output/system.xml'
    :type system: str, optional
    :param warming_steps: number of warming steps, defaults to 100000
    :type warming_steps: int, optional
    :param simulation_steps: number of simulation steps, defaults to 25000000
    :type simulation_steps: int, optional
    :param seed: random seed for reproducibility, defaults to None
    :type seed: int, optional
    :raises ValueError: different checks are performed and expections are raised if some of the fail.
    """

    # Temperature annealing
    Tstart = 5
    Tend = 300
    Tstep = 5
    
    simulation_format = simulation_format.upper()
    openmm_flag = simulation_format == "OPENMM"
    total_steps = warming_steps + simulation_steps

    if simulation_format not in ['OPENMM', 'AMBER', 'GROMACS', 'CHARMM']:
        raise ValueError(f"Unknown simulation_format {simulation_format}. It must be one of 'OPENMM', 'AMBER', 'GROMACS', or 'CHARMM'.")
    
    if not openmm_flag:
        assert topology is not None and positions is not None, "If the simulation format specified is not OpenMM be sure to pass both topology and positions files"
        if simulation_format == "AMBER":
            positions = app.AmberInpcrdFile(positions)
            topology = app.AmberPrmtopFile(topology, periodicBoxVectors=positions.boxVectors)
        elif simulation_format == "GROMACS":
            positions = app.GromacsGroFile(positions)
            topology = app.GromacsTopFile(topology, periodicBoxVectors=positions.getPeriodicBoxVectors())
        elif simulation_format == "CHARMM":
            topology = app.CharmmPsfFile(topology)
            positions = app.CharmmCrdFile(positions)
        system = topology.createSystem(nonbondedMethod=app.PME,
                                       nonbondedCutoff=10*openmmunit.angstrom,
                                       switchDistance=9*openmmunit.angstrom,
                                       constraints=app.HBonds,
                                       removeCMMotion=True,
                                       hydrogenMass=3.0*openmmunit.amu)
    else:
        assert pdb is not None and system is not None, "If the simulation format specified is OpenMM be sure to pass both pdb file and system.xml"
        pdb = app.PDBFile(f'{results_path}/system.pdb')
        topology = pdb.topology
        positions = pdb.positions
        system = openmm.XmlSerializer.deserialize(open(f'{results_path}/system.xml').read())

    print('Selecting simulation platform')
    try:
        platform = openmm.Platform.getPlatformByName("CUDA")
        platform.setPropertyDefaultValue('DeterministicForces', 'true')
        platform.setPropertyDefaultValue('CudaPrecision', 'mixed')
        platform.setPropertyDefaultValue('CudaDeviceIndex', '0')
        print('Using GPU:CUDA')
    except: 
        try:
            platform = openmm.Platform.getPlatformByName("OpenCL")
            platform.setPropertyDefaultValue('DeterministicForces', 'true')
            platform.setPropertyDefaultValue('Precision', 'mixed')
            print('Using GPU:OpenCL')
        except:
            platform = openmm.Platform.getPlatformByName("CPU")
            print("Switching to CPU, no GPU available.")

    integrator = openmm.LangevinMiddleIntegrator(Tstart * openmmunit.kelvin,
                                          1 / openmmunit.picosecond,
                                          0.001 * openmmunit.picosecond)
    if seed is not None:
        integrator.setRandomNumberSeed(seed)
    
    if not openmm_flag:
        simulation = app.Simulation(topology.topology, system, integrator, platform)
    else:
        simulation = app.Simulation(topology, system, integrator, platform)
        
    print('Adding reporters to the simulation')
    #every 0.1ns
    simulation.reporters.append(app.StateDataReporter(os.path.join(results_path, "statistics.csv"), 25000, step=True, time=True,
                                                totalEnergy=True, potentialEnergy=True, kineticEnergy=True, 
                                                temperature=True, volume=True, density=True,
                                                progress=True, remainingTime=True, speed=True, totalSteps=total_steps))
    #every 0.1ns
    simulation.reporters.append(app.StateDataReporter(stdout, 25000, step=True, time=True,
                                                totalEnergy=True, potentialEnergy=True, kineticEnergy=True, 
                                                temperature=True, volume=True, density=True,
                                                progress=True, remainingTime=True, speed=True, totalSteps=total_steps, separator='\t'))
    
    #every 0.1ns
    simulation.reporters.append(app.DCDReporter(os.path.join(results_path, "trajectory.dcd"),
                                            reportInterval=25000, enforcePeriodicBox=None))

    
    #every 1ns
    simulation.reporters.append(app.CheckpointReporter(os.path.join(results_path,"simualtion.chk"), 250000)) 

    print("Setting positions for the simulation")
    if not openmm_flag:
        simulation.context.setPositions(positions.positions.value_in_unit(openmmunit.nanometer))
    else:
        simulation.context.setPositions(positions)

    print("Minimizing system's energy")
    simulation.minimizeEnergy()

    print(f'Heating system in NVT ensemble for {warming_steps*0.001/1000} ns')
    # Calculate the number of temperature steps
    nT = int((Tend - Tstart) / Tstep)

    # Set initial velocities and temperature
    simulation.context.setVelocitiesToTemperature(Tstart)
    
    # Warm up the system gradually
    for i in range(nT):
        temperature = Tstart + i * Tstep
        integrator.setTemperature(temperature)
        print(f"Temperature set to {temperature} K.")
        simulation.step(int(warming_steps / nT))

    # Increase the timestep for production simulations
    integrator.setStepSize(0.004 * openmmunit.picoseconds)

    print(f'Adding a Montecarlo Barostat to the system')
    system.addForce(openmm.MonteCarloBarostat(1 * openmmunit.bar, Tend * openmmunit.kelvin))
    simulation.context.reinitialize(preserveState=True)

    print(f"Running simulation in NPT ensemble for {simulation_steps*0.004/1000} ns")
    simulation.step(simulation_steps) #25000000 = 100ns

    return